﻿namespace KPGBase
{
    internal class Maze
    {
        /// <summary> Grid cells </summary>
        public Cell[,] cells;
        /// <summary> Random used for generation </summary>
        Random r;

        /// <summary>
        /// Generate maze
        /// </summary>
        /// <param name="m"></param>
        /// <param name="n"></param>
        public void GenerateMaze(int m, int n)
        {
            r = new Random();
            InitGrid(m, n);
            GenMazeStack();
            //GenMazeRec(0, 0);
        }

        /// <summary>
        /// Generate maze using the stack algorithm
        /// </summary>
        private void GenMazeStack()
        {
            // start with cell[0,0]
            Stack<Cell> stack = new Stack<Cell>();
            stack.Push(cells[0, 0]);
            cells[0, 0].visited = true;
            while (stack.Count > 0)
            {
                Cell c = stack.Pop();

                // any unvisited neighbours
                List<Cell> unvisited = GetUnvisitedNns(c);
                if (unvisited.Count > 0)
                {
                    // randomly choose one unvisited neighbor
                    int nIndex = r.Next(unvisited.Count);
                    Cell nn = unvisited[nIndex];

                    // remove wall
                    if (nn.row == c.row)
                        if (nn.col < c.col) c.west = false;
                        else nn.west = false;
                    else if (nn.col == c.col)
                        if (nn.row < c.row) c.north = false;
                        else nn.north = false;

                    // mark neighbor as visited
                    nn.visited = true;

                    // push current to stack
                    stack.Push(c);

                    // push neighbor to stack
                    stack.Push(nn);
                }
            }
        }

        /// <summary>
        /// Generate maze using the recursive algorithm
        /// </summary>
        /// <param name="row"> Row of current cell </param>
        /// <param name="col"> Column of the current cell </param>
        private void GenMazeRec(int row, int col)
        {
            // mark current cell as visited
            Cell curr = cells[row, col];
            curr.visited = true;
            // get unvisited neighbors
            List<Cell> nbs = GetUnvisitedNns(curr);
            // while cell has any unvisted neighbors
            while (nbs.Count != 0)
            {
                // get random neighbor
                int nIndex = r.Next(nbs.Count);
                Cell nn = nbs[nIndex];

                // remove wall
                if (nn.row == curr.row)
                    if (nn.col < curr.col) curr.west = false;
                    else nn.west = false;
                else if (nn.col == curr.col)
                    if (nn.row < curr.row) curr.north = false;
                    else nn.north = false;

                // recursive call
                GenMazeRec(nn.row, nn.col);
                // refresh unvisited neighbor list
                nbs = GetUnvisitedNns(curr);
            }
        }

        /// <summary>
        /// Initializes grid with unvisited cells
        /// </summary>
        /// <param name="m"> Number of rows in grid </param>
        /// <param name="n"> Number of columns in grid </param>
        private void InitGrid(int m, int n)
        {
            cells = new Cell[m, n];
            for (int i = 0; i < cells.GetLength(0); i++)
                for (int j = 0; j < cells.GetLength(1); j++)
                {
                    cells[i, j] = new Cell(i, j);
                }
        }

        /// <summary>
        /// Get unvisited neighbours of the cell c
        /// </summary>
        /// <param name="c"> Cell </param>
        /// <returns> List with unvisited neighbors </returns>
        private List<Cell> GetUnvisitedNns(Cell c)
        {
            int xn = c.row - 1; // north neighbor row index
            int xs = c.row + 1; // south neighbor row index
            int ye = c.col + 1; // east neighbor column index
            int yw = c.col - 1; // west neighbor column index

            // For each index test if it is a valid index for the grid, and test if the neighbor was visited
            List<Cell> unvisited = new List<Cell>();
            if (xn >= 0 && !cells[xn, c.col].visited)
                unvisited.Add(cells[xn, c.col]);
            if (xs < cells.GetLength(0) && !cells[xs, c.col].visited)
                unvisited.Add(cells[xs, c.col]);
            if (ye < cells.GetLength(1) && !cells[c.row, ye].visited)
                unvisited.Add(cells[c.row, ye]);
            if (yw >= 0 && !cells[c.row, yw].visited)
                unvisited.Add(cells[c.row, yw]);

            return unvisited;
        }

    }
}
